/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 * Copyright (C) 2011 ObjectFanatics, Ltd. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.agent;

import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.weaver.api.core.Weaver;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.core.exception.NoWeavingNeededException;
import jp.objectfanatics.assertion.weaver.impl.core.WeaverImpl;

import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

/**
 * Agent for weaving assertion functions into class file bytes.
 *
 * @author Makoto Sato
 * @author Hirokatsu Endo
 */
public class Agent implements ClassFileTransformer {

    // TODO 2011-03-31: 深読みされる可能性もあるので、特に意味がなければ static を外しましょう。実際に Agent を使ってテストするようなタイミングでこのTODOを実行しましょう。
    private static final Weaver weaver = new WeaverImpl ();

    /**
     * Creates and adds instance of this class to instrumentation.
     *
     * @param agentArgs       agent options
     * @param instrumentation {@link Instrumentation}
     */
    public static void premain (String agentArgs, @NotNull Instrumentation instrumentation) {
        instrumentation.addTransformer (new Agent ());
    }

    /**
     * Weaves assertion functions into class file bytes.
     */
    @Override
    public byte[] transform (ClassLoader loader, String className, Class<?> classBeingRedefined, ProtectionDomain protectionDomain, @NotNull byte[] classfileBuffer) throws IllegalClassFormatException {
        try {
            // TODO 2011-03-31 try の中は、「return weaver.weave (classfileBuffer);」 だけでOKだと思われ。
            // do nothing if it's not class loading.
            if (!isClassLoading (classBeingRedefined)) {
                return null;
            }
            // weaves assertion functions into class file bytes if needed.
            else {
                return weaver.weave (classfileBuffer);
            }
        }
        catch (NoWeavingNeededException e) {
            // TODO 2011-03-31 nullを返せばOK。
            return classfileBuffer;
        }
        catch (IllegalUseOfConstraintAnnotationException e) {
            // Show detailed description programmatically because java implementation doesn't show error message.
            System.err.println (e.getDetailedDescription ());
            throw new IllegalClassFormatException (e.getDetailedDescription ());
        }
        catch (Throwable e) {
            // Do printStackTrace programmatically because java implementation doesn't show error message.
            e.printStackTrace ();
            throw new IllegalClassFormatException (e.getMessage ());
        }
    }

    /**
     * Returns true if, and only if, classBeingRedefined is null.
     *
     * @param classBeingRedefined class passed as a classBeingRedefined parameter of {@link #transform(ClassLoader, String, Class, ProtectionDomain, byte[])}.
     *
     * @return true if classBeingRedefined is null, otherwise false
     */
    private boolean isClassLoading (Class<?> classBeingRedefined) {
        return classBeingRedefined == null;
    }
}
